@@ -608,29 +608,35 @@ ethi_drag_data_received (GtkWidget *canvas,
 			 ETableHeaderItem *ethi)
 {
 	int found = FALSE;
-	int count = e_table_header_count(ethi->eth);
-	int column = atoi(data->data);
-	int drop_col = ethi->drop_col;
+	int count;
+	int column;
+	int drop_col;
 	int i;
-	ethi->drop_col = -1;
 
-	if (column < 0)
-		return;
-	for (i = 0; i < count; i++) {
-		ETableCol *ecol = e_table_header_get_column (ethi->eth, i);
-		if (ecol->col_idx == column) {
-			e_table_header_move(ethi->eth, i, drop_col);
-			found = TRUE;
-			break;
-		}
-	}
-	if (!found) {
-		count = e_table_header_count(ethi->full_header);
-		for (i = 0; i < count; i++) {
-			ETableCol *ecol = e_table_header_get_column (ethi->full_header, i);
-			if (ecol->col_idx == column) {
-				e_table_header_add_column (ethi->eth, ecol, drop_col);
-				break;
+	if (data->data) {
+		count = e_table_header_count(ethi->eth);
+		column = atoi(data->data);
+		drop_col = ethi->drop_col;
+		ethi->drop_col = -1;
+
+		if (column >= 0) {
+			for (i = 0; i < count; i++) {
+				ETableCol *ecol = e_table_header_get_column (ethi->eth, i);
+				if (ecol->col_idx == column) {
+					e_table_header_move(ethi->eth, i, drop_col);
+					found = TRUE;
+					break;
+				}
+			}
+			if (!found) {
+				count = e_table_header_count(ethi->full_header);
+				for (i = 0; i < count; i++) {
+					ETableCol *ecol = e_table_header_get_column (ethi->full_header, i);
+					if (ecol->col_idx == column) {
+						e_table_header_add_column (ethi->eth, ecol, drop_col);
+						break;
+					}
+				}
 			}
 		}
 	}
@@ -1193,6 +1199,7 @@ ethi_popup_field_chooser(GtkWidget *widget, EthiHeaderInfo *info)
 	GtkWidget *etfcd = e_table_field_chooser_dialog_new();
 	gtk_object_set(GTK_OBJECT(etfcd),
 		       "full_header", info->ethi->full_header,
+		       "header", info->ethi->eth,
 		       "dnd_code", info->ethi->dnd_code,
 		       NULL);
 	gtk_widget_show(etfcd);
@@ -1270,15 +1277,15 @@ static EPopupMenu ethi_context_menu [] = {
 	{ N_("Sort Descending"),           NULL, GTK_SIGNAL_FUNC(ethi_popup_sort_descending), NULL, 2},
 	{ N_("Unsort"),                    NULL, GTK_SIGNAL_FUNC(ethi_popup_unsort),          NULL, 0},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 0},
-	{ N_("Group By This Field"),       NULL, GTK_SIGNAL_FUNC(ethi_popup_group_field),     NULL, 0},
-	{ N_("Group By Box"),              NULL, GTK_SIGNAL_FUNC(ethi_popup_group_box),       NULL, 1},
+	{ N_("Group By This Field"),       NULL, GTK_SIGNAL_FUNC(ethi_popup_group_field),     NULL, 16},
+	{ N_("Group By Box"),              NULL, GTK_SIGNAL_FUNC(ethi_popup_group_box),       NULL, 128},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 1},
 	{ N_("Remove This Column"),        NULL, GTK_SIGNAL_FUNC(ethi_popup_remove_column),   NULL, 8},
 	{ N_("Add a Column..."),           NULL, GTK_SIGNAL_FUNC(ethi_popup_field_chooser),   NULL, 0},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 1},
-	{ N_("Alignment"),                 NULL, GTK_SIGNAL_FUNC(ethi_popup_alignment),       NULL, 1},
+	{ N_("Alignment"),                 NULL, GTK_SIGNAL_FUNC(ethi_popup_alignment),       NULL, 128},
 	{ N_("Best Fit"),                  NULL, GTK_SIGNAL_FUNC(ethi_popup_best_fit),        NULL, 2},
-	{ N_("Format Columns..."),         NULL, GTK_SIGNAL_FUNC(ethi_popup_format_columns),  NULL, 1},
+	{ N_("Format Columns..."),         NULL, GTK_SIGNAL_FUNC(ethi_popup_format_columns),  NULL, 128},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 1},
 	{ N_("Customize Current View..."), NULL, GTK_SIGNAL_FUNC(ethi_popup_customize_view),  NULL, 4},
 	{ NULL, NULL, NULL, NULL, 0 }
@@ -1297,7 +1304,8 @@ ethi_header_context_menu (ETableHeaderItem *ethi, GdkEventButton *event)
 			  (col->sortable ? 0 : 2) +
 			  (ethi->table ? 0 : 4) + 
 			  ((e_table_header_count (ethi->eth) > 1) ? 0 : 8),
-			  0, info);
+			  ((e_table_sort_info_get_can_group (ethi->sort_info)) ? 0 : 16) +
+			  128, info);
 }
 
 static void
@@ -1317,6 +1325,10 @@ ethi_change_sort_state (ETableHeaderItem *ethi, gdouble x)
 	int found = FALSE;
 	
 	col = e_table_header_get_column (ethi->eth, ethi_find_col_by_x (ethi, x));
+
+	if (col == NULL)
+		return;
+
 	model_col = col->col_idx;
 	
 	length = e_table_sort_info_grouping_get_count(ethi->sort_info);
