@@ -1762,7 +1762,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 				eti->click_count ++;
 				eti->row_guess = row;
 
-				if ((!eti_editing(eti)) && e_table_model_is_cell_editable(eti->table_model, col, row)) {
+				if ((!eti_editing(eti)) && e_table_model_is_cell_editable(eti->table_model, cursor_col, row)) {
 					e_table_item_enter_edit (eti, col, row);
 				}
 
@@ -1967,6 +1967,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 					 model_to_view_row(eti, cursor_row), cursor_col, e, &return_val);
 			if ((!return_val) && eti->cursor_mode != E_CURSOR_LINE && cursor_col != view_to_model_col(eti, 0))
 				eti_cursor_move_left (eti);
+			return_val = 1;
 			break;
 
 		case GDK_Right:
@@ -1980,6 +1981,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 					 model_to_view_row(eti, cursor_row), cursor_col, e, &return_val);
 			if ((!return_val) && eti->cursor_mode != E_CURSOR_LINE && cursor_col != view_to_model_col(eti, eti->cols - 1))
 				eti_cursor_move_right (eti);
+			return_val = 1;
 			break;
 			
 		case GDK_Up:
@@ -2030,7 +2032,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 					       NULL);
 
 				if (cursor_col >= 0 && cursor_row >= 0 && return_val &&
-				    (!eti_editing(eti)) && e_table_model_is_cell_editable(eti->table_model, model_to_view_col(eti, cursor_col), model_to_view_row(eti, cursor_row))) {
+				    (!eti_editing(eti)) && e_table_model_is_cell_editable(eti->table_model, cursor_col, model_to_view_row (eti, cursor_row))) {
 					e_table_item_enter_edit (eti, model_to_view_col(eti, cursor_col), model_to_view_row(eti, cursor_row));
 				}
 				break;
@@ -2090,7 +2092,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 					gint col, row;
 					row = model_to_view_row(eti, cursor_row);
 					col = model_to_view_col(eti, cursor_col);
-					if (col != -1 && row != -1 && e_table_model_is_cell_editable(eti->table_model, col, row)) {
+					if (col != -1 && row != -1 && e_table_model_is_cell_editable(eti->table_model, cursor_col, row)) { 
 						e_table_item_enter_edit (eti, col, row);
 					}
 				}
