@@ -124,6 +124,9 @@ static gint e_table_drag_source_event_cb (GtkWidget      *widget,
 
 static gint et_focus (GtkContainer *container, GtkDirectionType direction);
 
+static void scroll_off (ETable *et);
+static void scroll_on (ETable *et, gboolean down);
+
 static void
 et_disconnect_model (ETable *et)
 {
@@ -168,6 +171,8 @@ et_destroy (GtkObject *object)
 		g_source_remove(et->reflow_idle_id);
 	et->reflow_idle_id = 0;
 
+	scroll_off (et);
+
 	gtk_object_unref (GTK_OBJECT (et->model));
 	gtk_object_unref (GTK_OBJECT (et->full_header));
 	gtk_object_unref (GTK_OBJECT (et->header));
@@ -203,38 +208,39 @@ e_table_init (GtkObject *object)
 
 	GTK_WIDGET_SET_FLAGS (e_table, GTK_CAN_FOCUS);
 
-	gtk_table->homogeneous = FALSE;
+	gtk_table->homogeneous                      = FALSE;
 
-	e_table->sort_info = NULL;
-	e_table->group_info_change_id = 0;
-	e_table->reflow_idle_id = 0;
+	e_table->sort_info                          = NULL;
+	e_table->group_info_change_id               = 0;
+	e_table->reflow_idle_id                     = 0;
+	e_table->scroll_idle_id               = 0;
 
-	e_table->alternating_row_colors = 1;
-	e_table->horizontal_draw_grid = 1;
-	e_table->vertical_draw_grid = 1;
-	e_table->draw_focus = 1;
-	e_table->cursor_mode = E_CURSOR_SIMPLE;
-	e_table->length_threshold = 200;
+	e_table->alternating_row_colors             = 1;
+	e_table->horizontal_draw_grid               = 1;
+	e_table->vertical_draw_grid                 = 1;
+	e_table->draw_focus                         = 1;
+	e_table->cursor_mode                        = E_CURSOR_SIMPLE;
+	e_table->length_threshold                   = 200;
 
-	e_table->need_rebuild = 0;
-	e_table->rebuild_idle_id = 0;
+	e_table->need_rebuild                       = 0;
+	e_table->rebuild_idle_id                    = 0;
 
-	e_table->horizontal_scrolling = FALSE;
+	e_table->horizontal_scrolling               = FALSE;
 
-	e_table->click_to_add_message = NULL;
+	e_table->click_to_add_message               = NULL;
 
-	e_table->drag_get_data_row = -1;
-	e_table->drag_get_data_col = -1;
-	e_table->drop_row = -1;
-	e_table->drop_col = -1;
-	e_table->site = NULL;
-	e_table->drag_source_button_press_event_id = 0;
+	e_table->drag_get_data_row                  = -1;
+	e_table->drag_get_data_col                  = -1;
+	e_table->drop_row                           = -1;
+	e_table->drop_col                           = -1;
+	e_table->site                               = NULL;
+	e_table->drag_source_button_press_event_id  = 0;
 	e_table->drag_source_motion_notify_event_id = 0;
 
-	e_table->sorter = NULL;
-	e_table->selection = e_table_selection_model_new();
-	e_table->cursor_loc = E_TABLE_CURSOR_LOC_NONE;
-	e_table->spec = NULL;
+	e_table->sorter                             = NULL;
+	e_table->selection                          = e_table_selection_model_new();
+	e_table->cursor_loc                         = E_TABLE_CURSOR_LOC_NONE;
+	e_table->spec                               = NULL;
 }
 
 /* Grab_focus handler for the ETable */
@@ -430,13 +436,17 @@ click_to_add_cursor_change (ETableClickToAdd *etcta, int row, int col, ETable *e
 static void
 group_cursor_change (ETableGroup *etg, int row, ETable *et)
 {
-	if (et->cursor_loc == E_TABLE_CURSOR_LOC_ETCTA && et->click_to_add) {
-		e_table_click_to_add_commit(E_TABLE_CLICK_TO_ADD(et->click_to_add));
-	}
+	ETableCursorLoc old_cursor_loc;
+
+	old_cursor_loc = et->cursor_loc;
+
 	et->cursor_loc = E_TABLE_CURSOR_LOC_TABLE;
 	gtk_signal_emit (GTK_OBJECT (et),
 			 et_signals [CURSOR_CHANGE],
 			 row);
+
+	if (old_cursor_loc == E_TABLE_CURSOR_LOC_ETCTA && et->click_to_add)
+		e_table_click_to_add_commit(E_TABLE_CLICK_TO_ADD(et->click_to_add));
 }
 
 static void
@@ -485,6 +495,7 @@ group_key_press (ETableGroup *etg, int row, int col, GdkEvent *event, ETable *et
 
 	switch (key->keyval) {
 	case GDK_Page_Down:
+	case GDK_KP_Page_Down:
 		vadj = gtk_layout_get_vadjustment (GTK_LAYOUT (et->table_canvas));
 		y = CLAMP(vadj->value + (2 * vadj->page_size - 50), 0, vadj->upper);
 		y -= vadj->value;
@@ -495,6 +506,7 @@ group_key_press (ETableGroup *etg, int row, int col, GdkEvent *event, ETable *et
 		return_val = 1;
 		break;
 	case GDK_Page_Up:
+	case GDK_KP_Page_Up:
 		vadj = gtk_layout_get_vadjustment (GTK_LAYOUT (et->table_canvas));
 		y = CLAMP(vadj->value - (vadj->page_size - 50), 0, vadj->upper);
 		y -= vadj->value;
@@ -585,7 +597,12 @@ et_build_groups (ETable *et)
 				       et->model,
 				       et->sort_info,
 				       0);
-	e_canvas_vbox_add_item(E_CANVAS_VBOX(et->canvas_vbox), GNOME_CANVAS_ITEM(et->group));
+	
+	if (et->use_click_to_add_end)
+		e_canvas_vbox_add_item_start(E_CANVAS_VBOX(et->canvas_vbox), GNOME_CANVAS_ITEM(et->group));
+	else
+		e_canvas_vbox_add_item(E_CANVAS_VBOX(et->canvas_vbox), GNOME_CANVAS_ITEM(et->group));
+
 	gnome_canvas_item_set(GNOME_CANVAS_ITEM(et->group),
 			      "alternating_row_colors", et->alternating_row_colors,
 			      "horizontal_draw_grid", et->horizontal_draw_grid,
@@ -750,6 +767,8 @@ e_table_setup_table (ETable *e_table, ETableHeader *full_header, ETableHeader *h
 		"spacing", 10.0,
 		NULL);
 
+	et_build_groups(e_table);
+
 	if (e_table->use_click_to_add) {
 		e_table->click_to_add = gnome_canvas_item_new (
 			GNOME_CANVAS_GROUP(e_table->canvas_vbox),
@@ -759,15 +778,19 @@ e_table_setup_table (ETable *e_table, ETableHeader *full_header, ETableHeader *h
 			"message", e_table->click_to_add_message,
 			NULL);
 
-		e_canvas_vbox_add_item (
-			E_CANVAS_VBOX(e_table->canvas_vbox),
-			e_table->click_to_add);
+		if (e_table->use_click_to_add_end)
+			e_canvas_vbox_add_item (
+				E_CANVAS_VBOX(e_table->canvas_vbox),
+				e_table->click_to_add);
+		else
+			e_canvas_vbox_add_item_start (
+				E_CANVAS_VBOX(e_table->canvas_vbox),
+				e_table->click_to_add);
+
 		gtk_signal_connect (
 			GTK_OBJECT (e_table->click_to_add), "cursor_change",
 			GTK_SIGNAL_FUNC(click_to_add_cursor_change), e_table);
 	}
-
-	et_build_groups(e_table);
 }
 
 static void
@@ -964,6 +987,7 @@ et_real_construct (ETable *e_table, ETableModel *etm, ETableExtras *ete,
 		ete = e_table_extras_new();
 
 	e_table->use_click_to_add = specification->click_to_add;
+	e_table->use_click_to_add_end = specification->click_to_add_end;
 	e_table->click_to_add_message = e_utf8_from_locale_string (gettext (specification->click_to_add_message));
 	e_table->alternating_row_colors = specification->alternating_row_colors;
 	e_table->horizontal_draw_grid = specification->horizontal_draw_grid;
@@ -1662,6 +1686,39 @@ e_table_drag_highlight (ETable *table,
 {
 	g_return_if_fail(table != NULL);
 	g_return_if_fail(E_IS_TABLE(table));
+
+	if (row != -1) {
+		int x, y, width, height;
+		if (col == -1) {
+			e_table_get_cell_geometry (table, row, 0, &x, &y, &width, &height);
+			x = 0;
+			width = GTK_WIDGET (table->table_canvas)->allocation.width;
+		} else {
+			e_table_get_cell_geometry (table, row, col, &x, &y, &width, &height);
+			x += GTK_LAYOUT(table->table_canvas)->hadjustment->value;
+		}
+		y += GTK_LAYOUT(table->table_canvas)->vadjustment->value;
+
+		if (table->drop_highlight == NULL) {
+			table->drop_highlight =
+				gnome_canvas_item_new (gnome_canvas_root (table->table_canvas),
+						       gnome_canvas_rect_get_type (),
+						       "fill_color", NULL,
+						       /*						       "outline_color", "black",
+						       "width_pixels", 1,*/
+						       "outline_color_gdk", &(GTK_WIDGET (table)->style->fg[GTK_STATE_NORMAL]),
+						       NULL);
+		}
+		gnome_canvas_item_set (table->drop_highlight,
+				       "x1", (double) x,
+				       "x2", (double) x + width - 1,
+				       "y1", (double) y,
+				       "y2", (double) y + height - 1,
+				       NULL);
+	} else {
+		gtk_object_destroy (GTK_OBJECT (table->drop_highlight));
+		table->drop_highlight = NULL;
+	}
 }
 
 void
@@ -1669,6 +1726,11 @@ e_table_drag_unhighlight (ETable *table)
 {
 	g_return_if_fail(table != NULL);
 	g_return_if_fail(E_IS_TABLE(table));
+
+	if (table->drop_highlight) {
+		gtk_object_destroy (GTK_OBJECT (table->drop_highlight));
+		table->drop_highlight = NULL;
+	}
 }
 
 void e_table_drag_dest_set   (ETable               *table,
@@ -1868,32 +1930,21 @@ et_drag_data_delete(GtkWidget *widget,
 			 context);
 }
 
-static void
-et_drag_leave(GtkWidget *widget,
-	      GdkDragContext *context,
-	      guint time,
-	      ETable *et)
-{
-	gtk_signal_emit (GTK_OBJECT (et),
-			 et_signals [TABLE_DRAG_LEAVE],
-			 et->drop_row,
-			 et->drop_col,
-			 context,
-			 time);
-	et->drop_row = -1;
-	et->drop_col = -1;
-}
-
 static gboolean
-et_drag_motion(GtkWidget *widget,
+do_drag_motion(ETable *et,
 	       GdkDragContext *context,
 	       gint x,
 	       gint y,
-	       guint time,
-	       ETable *et)
+	       guint time)
 {
 	gboolean ret_val;
 	int row, col;
+	GtkWidget *widget;
+
+	widget = GTK_WIDGET (et);
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
 
 	e_table_get_cell_at (et, x, y, &row, &col);
 
@@ -1916,6 +1967,112 @@ et_drag_motion(GtkWidget *widget,
 			 y,
 			 time,
 			 &ret_val);
+
+	return ret_val;
+}
+
+static gboolean
+scroll_timeout (gpointer data)
+{
+	ETable *et = data;
+	int dy;
+	GtkAdjustment *v;
+	double value;
+
+	if (et->scroll_down)
+		dy = 20;
+	else
+		dy = -20;
+
+	v = GTK_LAYOUT(et->table_canvas)->vadjustment;
+
+	value = v->value;
+
+	gtk_adjustment_set_value(v, CLAMP(v->value + dy, v->lower, v->upper - v->page_size));
+
+	if (v->value != value)
+		do_drag_motion(et,
+			       et->last_drop_context,
+			       et->last_drop_x,
+			       et->last_drop_y,
+			       et->last_drop_time);
+			       
+
+	return TRUE;
+}
+
+static void
+scroll_on (ETable *et, gboolean down)
+{
+	if (et->scroll_idle_id == 0 || down != et->scroll_down) {
+		if (et->scroll_idle_id != 0)
+			g_source_remove (et->scroll_idle_id);
+		et->scroll_down = down;
+		et->scroll_idle_id = g_timeout_add (100, scroll_timeout, et);
+	}
+}
+
+static void
+scroll_off (ETable *et)
+{
+	if (et->scroll_idle_id) {
+		g_source_remove (et->scroll_idle_id);
+		et->scroll_idle_id = 0;
+	}
+}
+
+static void
+et_drag_leave(GtkWidget *widget,
+	      GdkDragContext *context,
+	      guint time,
+	      ETable *et)
+{
+	gtk_signal_emit (GTK_OBJECT (et),
+			 et_signals [TABLE_DRAG_LEAVE],
+			 et->drop_row,
+			 et->drop_col,
+			 context,
+			 time);
+	et->drop_row = -1;
+	et->drop_col = -1;
+
+	scroll_off (et);
+}
+
+static gboolean
+et_drag_motion(GtkWidget *widget,
+	       GdkDragContext *context,
+	       gint x,
+	       gint y,
+	       guint time,
+	       ETable *et)
+{
+	gboolean ret_val;
+
+	et->last_drop_x = x;
+	et->last_drop_y = y;
+	et->last_drop_time = time;
+	et->last_drop_context = context;
+
+	ret_val = do_drag_motion (et,
+				  context,
+				  x,
+				  y,
+				  time);
+
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+
+	if (y < 20 || y > widget->allocation.height - 20) {
+		if (y < 20)
+			scroll_on (et, FALSE);
+		else
+			scroll_on (et, TRUE);
+	} else {
+		scroll_off (et);
+	}
+
 	return ret_val;
 }
 
@@ -1930,6 +2087,9 @@ et_drag_drop(GtkWidget *widget,
 	gboolean ret_val;
 	int row, col;
 
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+
 	e_table_get_cell_at (et, x, y, &row, &col);
 
 	if (row != et->drop_row && col != et->drop_row) {
@@ -1962,6 +2122,9 @@ et_drag_drop(GtkWidget *widget,
 			 &ret_val);
 	et->drop_row = -1;
 	et->drop_col = -1;
+
+	scroll_off (et);
+
 	return ret_val;
 }
 
@@ -1977,6 +2140,9 @@ et_drag_data_received(GtkWidget *widget,
 {
 	int row, col;
 
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+
 	e_table_get_cell_at (et, x, y, &row, &col);
 
 	gtk_signal_emit (GTK_OBJECT (et),
@@ -2043,7 +2209,6 @@ e_table_drag_source_event_cb (GtkWidget      *widget,
 							      site->actions,
 							      i, event);
 
-
 				info = g_dataset_get_data (context, "gtk-info");
 
 				if (!info->icon_window) {
