@@ -90,6 +90,8 @@ struct ETreePriv {
 	ETableSpecification *spec;
 
 	int reflow_idle_id;
+	int scroll_idle_id;
+	int hover_idle_id;
 
 	int table_model_change_id;
 	int table_row_change_id;
@@ -117,12 +119,23 @@ struct ETreePriv {
 
 	guint horizontal_scrolling : 1;
 
+	guint scroll_down : 1;
+
 	ECursorMode cursor_mode;
 
 	int drop_row;
 	ETreePath drop_path;
 	int drop_col;
 
+	GnomeCanvasItem *drop_highlight;
+	int last_drop_x;
+	int last_drop_y;
+	int last_drop_time;
+	GdkDragContext *last_drop_context;
+
+	int hover_x;
+	int hover_y;
+
 	int drag_row;
 	ETreePath drag_path;
 	int drag_col;
@@ -182,6 +195,11 @@ static gint e_tree_drag_source_event_cb (GtkWidget      *widget,
 
 static gint et_focus (GtkContainer *container, GtkDirectionType direction);
 
+static void scroll_off (ETree *et);
+static void scroll_on (ETree *et, gboolean down);
+static void hover_off (ETree *et);
+static void hover_on (ETree *et, int x, int y);
+
 static void
 et_disconnect_from_etta (ETree *et)
 {
@@ -217,6 +235,9 @@ et_destroy (GtkObject *object)
 		g_source_remove(et->priv->reflow_idle_id);
 	et->priv->reflow_idle_id = 0;
 
+	scroll_off (et);
+	hover_off (et);
+
 	et_disconnect_from_etta (et);
 
 	gtk_object_unref (GTK_OBJECT (et->priv->etta));
@@ -241,6 +262,7 @@ et_destroy (GtkObject *object)
 	gtk_widget_destroy (GTK_WIDGET (et->priv->table_canvas));
 
 	g_free(et->priv);
+	et->priv = NULL;
 
 	(*parent_class->destroy)(object);
 }
@@ -248,8 +270,8 @@ et_destroy (GtkObject *object)
 static void
 e_tree_init (GtkObject *object)
 {
-	ETree *e_tree = E_TREE (object);
-	GtkTable *gtk_table = GTK_TABLE (object);
+	ETree *e_tree                                    = E_TREE (object);
+	GtkTable *gtk_table                              = GTK_TABLE (object);
 
 	GTK_WIDGET_SET_FLAGS (e_tree, GTK_CAN_FOCUS);
 
@@ -267,8 +289,10 @@ e_tree_init (GtkObject *object)
 	e_tree->priv->sort_info                          = NULL;
 	e_tree->priv->sorter                             = NULL;
 	e_tree->priv->reflow_idle_id                     = 0;
+	e_tree->priv->scroll_idle_id                     = 0;
+	e_tree->priv->hover_idle_id                     = 0;
 
-	e_tree->priv->alternating_row_colors		 = 1;
+	e_tree->priv->alternating_row_colors             = 1;
 	e_tree->priv->horizontal_draw_grid               = 1;
 	e_tree->priv->vertical_draw_grid                 = 1;
 	e_tree->priv->draw_focus                         = 1;
@@ -281,6 +305,15 @@ e_tree_init (GtkObject *object)
 	e_tree->priv->drop_row                           = -1;
 	e_tree->priv->drop_path                          = NULL;
 	e_tree->priv->drop_col                           = -1;
+	e_tree->priv->drop_highlight                     = NULL;
+
+	e_tree->priv->last_drop_x                        = 0;
+	e_tree->priv->last_drop_y                        = 0;
+	e_tree->priv->last_drop_time                     = 0;
+	e_tree->priv->last_drop_context                  = NULL;
+
+	e_tree->priv->hover_x                            = 0;
+	e_tree->priv->hover_y                            = 0;
 
 	e_tree->priv->drag_row                           = -1;
 	e_tree->priv->drag_path                          = NULL;
@@ -534,6 +567,7 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 
 	switch (key->keyval) {
 	case GDK_Page_Down:
+	case GDK_KP_Page_Down:
 		vadj = gtk_layout_get_vadjustment (GTK_LAYOUT (et->priv->table_canvas));
 		y = CLAMP(vadj->value + (2 * vadj->page_size - 50), 0, vadj->upper);
 		y -= vadj->value;
@@ -544,6 +578,7 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 		return_val = 1;
 		break;
 	case GDK_Page_Up:
+	case GDK_KP_Page_Up:
 		vadj = gtk_layout_get_vadjustment (GTK_LAYOUT (et->priv->table_canvas));
 		y = CLAMP(vadj->value - (vadj->page_size - 50), 0, vadj->upper);
 		y -= vadj->value;
@@ -555,6 +590,7 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 		break;
 	case '=':
 	case GDK_Right:
+	case GDK_KP_Right:
 		if (row != -1) {
 			path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
 			if (path)
@@ -564,6 +600,7 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 		break;
 	case '-':
 	case GDK_Left:
+	case GDK_KP_Left:
 		if (row != -1) {
 			path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
 			if (path)
@@ -718,9 +755,10 @@ e_tree_set_state_object(ETree *e_tree, ETableState *state)
 	if (e_tree->priv->sort_info)
 		gtk_object_unref(GTK_OBJECT(e_tree->priv->sort_info));
 
-	if (state->sort_info)
+	if (state->sort_info) {
 		e_tree->priv->sort_info = e_table_sort_info_duplicate(state->sort_info);
-	else
+		e_table_sort_info_set_can_group (e_tree->priv->sort_info, FALSE);
+	} else
 		e_tree->priv->sort_info = NULL;
 
 	if (e_tree->priv->header_item)
@@ -810,7 +848,8 @@ e_tree_get_state_object (ETree *e_tree)
 
 	state = e_table_state_new();
 	state->sort_info = e_tree->priv->sort_info;
-	gtk_object_ref(GTK_OBJECT(state->sort_info));
+	if (state->sort_info)
+		gtk_object_ref(GTK_OBJECT(state->sort_info));
 
 	state->col_count = e_table_header_count (e_tree->priv->header);
 	full_col_count = e_table_header_count (e_tree->priv->full_header);
@@ -936,6 +975,8 @@ et_real_construct (ETree *e_tree, ETreeModel *etm, ETableExtras *ete,
 	e_tree->priv->sort_info = state->sort_info;
 	gtk_object_ref (GTK_OBJECT (e_tree->priv->sort_info));
 
+	e_table_sort_info_set_can_group (e_tree->priv->sort_info, FALSE);
+
 	gtk_object_set(GTK_OBJECT(e_tree->priv->header),
 		       "sort_info", e_tree->priv->sort_info,
 		       NULL);
@@ -1591,6 +1632,29 @@ e_tree_drag_get_data (ETree         *tree,
 
 }
 
+#if 0
+static void
+e_tree_request_hightlight_redraw (ETree *tree)
+{
+	int row = tree->drop_highlight_row;
+	int col = tree->drop_highlight_col;
+
+	if (row != -1) {
+		int x, y, width, height;
+		if (col == -1) {
+			e_tree_get_cell_geometry (tree, row, 0, &x, &y, &width, &height);
+			x = 0;
+			width = tree->allocation.width;
+		} else {
+			e_tree_get_cell_geometry (tree, row, col, &x, &y, &width, &height);
+			x += GTK_LAYOUT(tree->priv->table_canvas)->hadjustment->value;
+		}
+		y += GTK_LAYOUT(tree->priv->table_canvas)->vadjustment->value;
+		gnome_canvas_request_redraw (tree->priv->table_canvas, x, y, x + width - 1, y + height - 1);
+	}
+}
+#endif
+
 /**
  * e_tree_drag_highlight:
  * @tree:
@@ -1598,14 +1662,48 @@ e_tree_drag_get_data (ETree         *tree,
  * @col:
  *
  * Set col to -1 to highlight the entire row.
+ * Set row to -1 to turn off the highlight.
  */
 void
 e_tree_drag_highlight (ETree *tree,
-			int     row,
-			int     col)
+		       int     row,
+		       int     col)
 {
 	g_return_if_fail(tree != NULL);
 	g_return_if_fail(E_IS_TREE(tree));
+
+	if (row != -1) {
+		int x, y, width, height;
+		if (col == -1) {
+			e_tree_get_cell_geometry (tree, row, 0, &x, &y, &width, &height);
+			x = 0;
+			width = GTK_WIDGET (tree->priv->table_canvas)->allocation.width;
+		} else {
+			e_tree_get_cell_geometry (tree, row, col, &x, &y, &width, &height);
+			x += GTK_LAYOUT(tree->priv->table_canvas)->hadjustment->value;
+		}
+		y += GTK_LAYOUT(tree->priv->table_canvas)->vadjustment->value;
+
+		if (tree->priv->drop_highlight == NULL) {
+			tree->priv->drop_highlight =
+				gnome_canvas_item_new (gnome_canvas_root (tree->priv->table_canvas),
+						       gnome_canvas_rect_get_type (),
+						       "fill_color", NULL,
+						       /*						       "outline_color", "black",
+						       "width_pixels", 1,*/
+						       "outline_color_gdk", &(GTK_WIDGET (tree)->style->fg[GTK_STATE_NORMAL]),
+						       NULL);
+		}
+		gnome_canvas_item_set (tree->priv->drop_highlight,
+				       "x1", (double) x,
+				       "x2", (double) x + width - 1,
+				       "y1", (double) y,
+				       "y2", (double) y + height - 1,
+				       NULL);
+	} else {
+		gtk_object_destroy (GTK_OBJECT (tree->priv->drop_highlight));
+		tree->priv->drop_highlight = NULL;
+	}
 }
 
 void
@@ -1613,6 +1711,11 @@ e_tree_drag_unhighlight (ETree *tree)
 {
 	g_return_if_fail(tree != NULL);
 	g_return_if_fail(E_IS_TREE(tree));
+
+	if (tree->priv->drop_highlight) {
+		gtk_object_destroy (GTK_OBJECT (tree->priv->drop_highlight));
+		tree->priv->drop_highlight = NULL;
+	}
 }
 
 void e_tree_drag_dest_set   (ETree               *tree,
@@ -1889,40 +1992,28 @@ et_drag_data_delete(GtkWidget *widget,
 			 context);
 }
 
-static void
-et_drag_leave(GtkWidget *widget,
-	      GdkDragContext *context,
-	      guint time,
-	      ETree *et)
-{
-	gtk_signal_emit (GTK_OBJECT (et),
-			 et_signals [TREE_DRAG_LEAVE],
-			 et->priv->drop_row,
-			 et->priv->drop_path,
-			 et->priv->drop_col,
-			 context,
-			 time);
-	et->priv->drop_row = -1;
-	et->priv->drop_col = -1;
-}
-
 static gboolean
-et_drag_motion(GtkWidget *widget,
+do_drag_motion(ETree *et,
 	       GdkDragContext *context,
 	       gint x,
 	       gint y,
-	       guint time,
-	       ETree *et)
+	       guint time)
 {
-	gboolean ret_val;
+	gboolean ret_val = FALSE;
 	int row, col;
 	ETreePath path;
+	GtkWidget *widget;
+
+	widget = GTK_WIDGET (et);
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
 	e_tree_get_cell_at (et,
 			    x,
 			    y,
 			    &row,
 			    &col);
-	if (row != et->priv->drop_row && col != et->priv->drop_row) {
+	if (row != et->priv->drop_row && col != et->priv->drop_col) {
 		gtk_signal_emit (GTK_OBJECT (et),
 				 et_signals [TREE_DRAG_LEAVE],
 				 et->priv->drop_row,
@@ -1948,6 +2039,167 @@ et_drag_motion(GtkWidget *widget,
 			 y,
 			 time,
 			 &ret_val);
+
+	return ret_val;
+}
+
+static gboolean
+scroll_timeout (gpointer data)
+{
+	ETree *et = data;
+	int dy;
+	GtkAdjustment *v;
+	double value;
+
+	if (et->priv->scroll_down)
+		dy = 20;
+	else
+		dy = -20;
+
+	v = GTK_LAYOUT(et->priv->table_canvas)->vadjustment;
+
+	value = v->value;
+
+	gtk_adjustment_set_value(v, CLAMP(v->value + dy, v->lower, v->upper - v->page_size));
+
+	if (v->value != value)
+		do_drag_motion(et,
+			       et->priv->last_drop_context,
+			       et->priv->last_drop_x,
+			       et->priv->last_drop_y,
+			       et->priv->last_drop_time);
+			       
+
+	return TRUE;
+}
+
+static void
+scroll_on (ETree *et, gboolean down)
+{
+	if (et->priv->scroll_idle_id == 0 || down != et->priv->scroll_down) {
+		if (et->priv->scroll_idle_id != 0)
+			g_source_remove (et->priv->scroll_idle_id);
+		et->priv->scroll_down = down;
+		et->priv->scroll_idle_id = g_timeout_add (100, scroll_timeout, et);
+	}
+}
+
+static void
+scroll_off (ETree *et)
+{
+	if (et->priv->scroll_idle_id) {
+		g_source_remove (et->priv->scroll_idle_id);
+		et->priv->scroll_idle_id = 0;
+	}
+}
+ 
+static gboolean
+hover_timeout (gpointer data)
+{
+	ETree *et = data;
+	GtkWidget *widget = data;
+	int x = et->priv->hover_x;
+	int y = et->priv->hover_y;
+	int row, col;
+	ETreePath path;
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+	e_tree_get_cell_at (et,
+			    x,
+			    y,
+			    &row,
+			    &col);
+
+	path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
+	if (path && e_tree_model_node_is_expandable (E_TREE_MODEL (et->priv->sorted), path)) {
+		e_tree_table_adapter_node_set_expanded (et->priv->etta, path, TRUE);
+	}
+
+	return TRUE;
+}
+
+static void
+hover_on (ETree *et, int x, int y)
+{
+	et->priv->hover_x = x;
+	et->priv->hover_y = y;
+	if (et->priv->hover_idle_id != 0)
+		g_source_remove (et->priv->hover_idle_id);
+	et->priv->hover_idle_id = g_timeout_add (500, hover_timeout, et);
+}
+
+static void
+hover_off (ETree *et)
+{
+	if (et->priv->hover_idle_id) {
+		g_source_remove (et->priv->hover_idle_id);
+		et->priv->hover_idle_id = 0;
+	}
+}
+
+static void
+et_drag_leave(GtkWidget *widget,
+	      GdkDragContext *context,
+	      guint time,
+	      ETree *et)
+{
+	gtk_signal_emit (GTK_OBJECT (et),
+			 et_signals [TREE_DRAG_LEAVE],
+			 et->priv->drop_row,
+			 et->priv->drop_path,
+			 et->priv->drop_col,
+			 context,
+			 time);
+	et->priv->drop_row = -1;
+	et->priv->drop_col = -1;
+
+	scroll_off (et);
+	hover_off (et);
+}
+
+static gboolean
+et_drag_motion(GtkWidget *widget,
+	       GdkDragContext *context,
+	       gint x,
+	       gint y,
+	       guint time,
+	       ETree *et)
+{
+	int ret_val;
+
+	et->priv->last_drop_x = x;
+	et->priv->last_drop_y = y;
+	et->priv->last_drop_time = time;
+	et->priv->last_drop_context = context;
+
+	if (et->priv->hover_idle_id != 0) {
+		if (abs (et->priv->hover_x - x) > 3 ||
+		    abs (et->priv->hover_y - y) > 3) {
+			hover_on (et, x, y);
+		}
+	} else {
+		hover_on (et, x, y);
+	}
+
+	ret_val = do_drag_motion (et,
+				  context,
+				  x,
+				  y,
+				  time);
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+
+	if (y < 20 || y > widget->allocation.height - 20) {
+		if (y < 20)
+			scroll_on (et, FALSE);
+		else
+			scroll_on (et, TRUE);
+	} else {
+		scroll_off (et);
+	}
+
 	return ret_val;
 }
 
@@ -1959,9 +2211,11 @@ et_drag_drop(GtkWidget *widget,
 	     guint time,
 	     ETree *et)
 {
-	gboolean ret_val;
+	gboolean ret_val = FALSE;
 	int row, col;
 	ETreePath path;
+	y -= widget->allocation.y;
+	x -= widget->allocation.x;
 	e_tree_get_cell_at(et,
 			   x,
 			   y,
@@ -2005,6 +2259,8 @@ et_drag_drop(GtkWidget *widget,
 	et->priv->drop_row = -1;
 	et->priv->drop_path = NULL;
 	et->priv->drop_col = -1;
+
+	scroll_off (et);
 	return ret_val;
 }
 
@@ -2020,6 +2276,8 @@ et_drag_data_received(GtkWidget *widget,
 {
 	int row, col;
 	ETreePath path;
+	y -= widget->allocation.y;
+	x -= widget->allocation.x;
 	e_tree_get_cell_at(et,
 			   x,
 			   y,
@@ -2092,7 +2350,6 @@ e_tree_drag_source_event_cb (GtkWidget      *widget,
 							     site->actions,
 							     i, event);
 
-
 				info = g_dataset_get_data (context, "gtk-info");
 
 				if (!info->icon_window) {
