@@ -48,13 +48,14 @@ e_completion_match_construct (ECompletionMatch *match)
 	g_return_if_fail (match != NULL);
 	
 	match->match_text = NULL;
-	match->menu_text = NULL;
-	match->score = 0;
+	match->menu_text  = NULL;
+	match->score      = 0;
 	match->sort_major = 0;
 	match->sort_minor = 0;
-	match->user_data = NULL;
-	match->ref = 1;
-	match->destroy = NULL;
+	match->user_data  = NULL;
+	match->ref        = 1;
+	match->hit_count  = 0;
+	match->destroy    = NULL;
 }
 
 void
@@ -84,10 +85,13 @@ e_completion_match_set_text (ECompletionMatch *match,
 			     const gchar *match_text,
 			     const gchar *menu_text)
 {
+	gchar *to_be_freed_match_text;
+	gchar *to_be_freed_menu_text;
+
 	g_return_if_fail (match != NULL);
 
-	g_free (match->match_text);
-	g_free (match->menu_text);
+	to_be_freed_match_text = match->match_text;
+	to_be_freed_menu_text  = match->menu_text;
 
 	if (match_text == NULL) {
 		match_text = "Unknown_Match";
@@ -103,6 +107,9 @@ e_completion_match_set_text (ECompletionMatch *match,
 
 	match->match_text = g_strdup (match_text);
 	match->menu_text  = g_strdup (menu_text);
+
+	g_free (to_be_freed_match_text);
+	g_free (to_be_freed_menu_text);
 }
 
 const gchar *
