@@ -158,7 +158,7 @@ g_utf8_find_prev_char (const char *str,
 {
   utf8_check (str, -1);
 
-  for (--p; p > str; --p)
+  for (--p; p >= str; --p)
     {
       if ((*p & 0xc0) != 0x80)
 	return (gchar *)p;
