@@ -540,7 +540,7 @@ e_categories_init (ECategories *categories)
 	gtk_object_unref(GTK_OBJECT (ecml));
 }
 
-void
+static void
 e_categories_destroy (GtkObject *object)
 {
 	ECategories *categories = E_CATEGORIES(object);
@@ -566,6 +566,9 @@ e_categories_destroy (GtkObject *object)
 	g_free(categories->priv->selected_list);
 	g_free(categories->priv);
 	categories->priv = NULL;
+
+	if (GTK_OBJECT_CLASS (parent_class)->destroy)
+		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
 }
 
 /**
